/* 
This file is part of project maximizing the number of directed P_4s
 in T_3-free graphs.
 
               I. Choi B. Lidicky, F. Pfender
 
It provides an upper bound for the program (P) from Claim ??.

 
Usage:
 g++ mesh-opt.cpp -Wall -O3 -o mesh-opt && ./mesh-opt

If you have OpenMP, you could use
 g++ mesh-opt.cpp -Wall -O3 -o mesh-opt -fopenmp && ./mesh-opt
 
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <istream>
#include <vector>
#include <utility>
#include <assert.h>
#include <cstring>
#include <algorithm>
#include <cstdarg>
#include <cmath>
#include <limits> 


using namespace std;


const double max21 = 0.21;
const int steps = 100; // Number of steps for each Xi in the recursion
const double stepsize = max21/steps;


const double Xmax = max21 + 2*stepsize; // Upper bound for Xi

const double funkyDegree = 0.17;

const int funkyDegreeInt = (funkyDegree*steps)/Xmax - 4; // the - 4 is to cover all rounding errors 

const double extra =  0.5/steps;

int main(int argc, char *argv[])
{

	double max_total =  0;

// Configruation 1
    const int o_max[] = {0,  steps, 0, steps, 0, 0};
    const int i_max[] = {0,  steps, 0, steps, 0, 0};

// Configruation 2
//    const int o_max[] = {0,  steps, 0, 0, steps, 0};
//    const int i_max[] = {0,  steps, 0, 0, steps, 0};

// Configruation 3
//    const int o_max[] = {0,  steps, 0, steps, 0, 0};
//    const int i_max[] = {0,  steps, 0, 0, steps, 0};
    
// Configruation 4
//    const int o_max[] = {0,  steps,     0, 0, steps, 0};
//    const int i_max[] = {0,      0, steps, 0, steps, 0};

    
    
#pragma omp parallel for ordered schedule(dynamic)
	for (int o1 = 1; o1 <= o_max[1]; o1++)
    {
        double max_inside = 0;
        int o[6];
        int i[6];
        int n[6];
        o[1] = o1;
        
        for (i[1] = 0; i[1] <= i_max[1]; i[1]++)
        for (i[2] = 0; i[2] <= i_max[2]; i[2]++)
        for (i[3] = 0; i[3] <= i_max[3]; i[3]++)
	    for (i[4] = 0; i[4] <= i_max[4]; i[4]++)
	    for (i[5] = 0; i[5] <= i_max[5]; i[5]++)

        for (o[2] = 0; o[2] <= max(0,o_max[2]-i[2]); o[2]++)
        for (o[3] = 0; o[3] <= max(0,o_max[3]-i[3]); o[3]++)
        for (o[4] = 0; o[4] <= max(0,o_max[4]-i[4]); o[4]++)
        for (o[5] = 0; o[5] <= max(0,o_max[5]-i[5]); o[5]++)
        
	{
		for (int j = 1; j <= 5; ++j) n[j] = steps - i[j] - o[j];

        
		// Kill entires with too few funky edges
    if (
	(o[1]+i[1]+n[2]+i[2]+i[3]+o[3]+i[4]+o[4]+n[5]+o[5] < funkyDegreeInt) ||
	(o[2]+i[2]+n[3]+i[3]+i[4]+o[4]+i[5]+o[5]+n[1]+o[1] < funkyDegreeInt) ||
	(o[3]+i[3]+n[4]+i[4]+i[5]+o[5]+i[1]+o[1]+n[2]+o[2] < funkyDegreeInt) ||
	(o[4]+i[4]+n[5]+i[5]+i[1]+o[1]+i[2]+o[2]+n[3]+o[3] < funkyDegreeInt) ||
	(o[5]+i[5]+n[1]+i[1]+i[2]+o[2]+i[3]+o[3]+n[4]+o[4] < funkyDegreeInt)
	) continue;

		int valueInt =
             o[1]*n[2]*n[3]+i[1]*o[3]*n[4]+n[3]*i[4]*o[1]+n[1]*n[2]*i[3] +
             o[2]*n[3]*n[4]+i[2]*o[4]*n[5]+n[4]*i[5]*o[2]+n[2]*n[3]*i[4] +
             o[3]*n[4]*n[5]+i[3]*o[5]*n[1]+n[5]*i[1]*o[3]+n[3]*n[4]*i[5] +
             o[4]*n[5]*n[1]+i[4]*o[1]*n[2]+n[1]*i[2]*o[4]+n[4]*n[5]*i[1] +
             o[5]*n[1]*n[2]+i[5]*o[2]*n[3]+n[2]*i[3]*o[5]+n[5]*n[1]*i[2];
        double value = valueInt*Xmax*Xmax*Xmax/(double)(steps*steps*steps);

		if (max_inside < value)
		{	
			max_inside = value;
			//cout << value+extra << " " << value << endl;
		}
	}
#pragma omp ordered
        {        
        cerr << "Done " << o1 << "/" << steps << endl;
        if (max_inside > max_total)
        {
            max_total = max_inside;
            cout << "Max(P)+extra=" << max_inside+extra << " Max(P)=" << max_inside << endl;
        }
        }
    }
	
    cout << "Discretization of (P)  is bounded by " << max_total << " and (P) is bounded by " << max_total + extra << endl;
    
	return 0;
}

